#    Quizpy - Creating Moodle exams in Python
#    Copyright (C) 2021  Sebastian Bräuer
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

import xml.etree.ElementTree as ET
import typing

from dataclasses import dataclass, field

from .common import Question


@dataclass
class Choice:
    """Represents an answer option in a Multiple Choice question or a Short Answer question.

    :param text: The source text to be rendered for this answer option.
    :type text: str

    :param fraction: The fraction of the total points to be given for this answer as percentage between 0 and 100.
    :type fraction: float

    :param feedback: Specific feedback for choosing this answer option
    :type feedback: str

    :param format: format for rendering both answer text and feedback.
        Can either be `html` (default), `markdown`, `plain_text` or `moodle_auto_format`.
    """
    text: str
    fraction: float  # TODO: Should this be int?

    feedback: str = ''
    format: str = 'html'

    def to_xml(self) -> ET.Element:
        answer = ET.Element('answer')
        answer.set('fraction', str(self.fraction))
        answer.set('format', self.format)

        ET.SubElement(answer, 'text').text = self.text
        fb = ET.SubElement(answer, 'feedback')
        fb.set('format', self.format)
        ET.SubElement(fb, 'text').text = self.feedback

        return answer

    def to_cloze(self) -> str:
        """Renders a cloze version of this choice.
        This method is called by the applicable question classes and should not be used directly.

        :return: A Cloze-representation of this choice.
        :rtype: str
        """
        if self.fraction == 100.0:
            s = f"={self.text}#{self.feedback}"
        elif self.fraction == 0.0:
            s = f"{self.text}#{self.feedback}"
        else:
            s = f"%{int(self.fraction)}%{self.text}#{self.feedback}"
        return s


@dataclass
class MultipleChoice(Question):
    """Standard Multiple choice question.

    :param shuffle: If `True` then answers will be shuffled every time they are presented in the quiz.
    :type shuffle: bool

    :param single: If `True` then only a single answer can be selected.
    :type single: bool

    :param numbering: Number answers with lowercase letters (`"abc"`), uppercase (`"ABC"`) or with integers (`"123"`)
    :type numbering: str

    :param correct_feedback: Feedback given for a fully correct answer (100%).
    :type correct_feedback: str

    :param partial_feedback: Feedback given for a partially correct answer (<100%).
    :type partial_feedback: str

    :param incorrect_feedback: Feedback given for a wrong answer (0%).
    :type incorrect_feedback: str
    """
    shuffle: bool = True
    single: bool = False
    numbering: str = 'abc'
    choices: typing.List[Choice] = field(default_factory=list)

    correct_feedback: str = 'Die Antwort ist richtig.'
    partial_feedback: str = 'Die Antwort ist teilweise richtig.'
    incorrect_feedback: str = 'Die Antwort ist falsch.'

    def add_choice(self, text: str, fraction: float, feedback: str = '', format: str = 'html'):
        """Convenience method to add a choice to the question.

        :param text: The source text to be rendered for this answer option.
        :type text: str

        :param fraction: The fraction of the total points to be given for this answer as percentage between 0 and 100.
        :type fraction: float

        :param feedback: Specific feedback for choosing this answer option
        :type feedback: str

        :param format: format for rendering both answer text and feedback.
            Can either be `html` (default), `markdown`, `plain_text` or `moodle_auto_format`.

        :returns: The newly added choice.
        :rtype: Choice
        """

        c = Choice(text, fraction,feedback,format)
        self.choices.append(c)
        return c

    def to_xml(self):
        node = self.generate_common_xml('multichoice')
        node.set('type', 'multichoice')

        ET.SubElement(node, 'shuffleanswers').text = 'true' if self.shuffle else 'false'

        ET.SubElement(node, 'single').text = 'true' if self.single else 'false'
        ET.SubElement(node, 'answernumbering').text = self.numbering

        correct_feedback = ET.SubElement(node, 'correctfeedback')
        correct_feedback.set('format', self.format)
        ET.SubElement(correct_feedback, 'text').text = self.correct_feedback

        partial_feedback = ET.SubElement(node, 'partiallycorrectfeedback')
        partial_feedback.set('format', self.format)
        ET.SubElement(partial_feedback, 'text').text = self.partial_feedback

        wrong_feedback = ET.SubElement(node, 'incorrectfeedback')
        wrong_feedback.set('format', self.format)
        ET.SubElement(wrong_feedback, 'text').text = self.incorrect_feedback

        for c in self.choices:
            node.append(c.to_xml())

        return node

    def to_cloze(self, weight: int, display_type: str = "dropdown", alignment: str = 'horizontal', shuffle: bool = False) -> str:
        """Renders a MC question in Cloze format to be used as an inline question.

        :param weight: The weight of the question in the total sum of all inline questions.
        :type weight: int

        :param display_type: Determines how the question is rendered.
            Can either be `"dropdown"`, `"radio"` or `"checkbox"`.
        :type display_type: str

        :param alignment: If the `display_type` is "radio" or "checkbox" then you can choose between "horizontal" and
            "vertical" display of the radio buttons / checkboxes.
        :type alignment: str

        :param shuffle: If `True` the choices will be shuffled.
        :type shuffle: bool
        """

        if display_type not in {'dropdown', 'radio', 'checkbox'}:
            raise ValueError("display_type must be either 'dropdown', 'radio' or 'checkbox'!")

        if alignment not in {'horizontal', 'vertical'}:
            raise ValueError("alignment must be either 'horizontal', 'vertical'!")

        if display_type == 'dropdown':
            cloze_type = 'MC'
        elif display_type == 'radio':
            if alignment == 'horizontal':
                cloze_type = 'MCH'
            else:
                cloze_type = 'MCV'
        else:
            if alignment == 'horizontal':
                cloze_type = 'MRH'
            else:
                cloze_type = 'MR'

        if shuffle:
            cloze_type += 'S'

        cloze = f'{weight}:{cloze_type}:'
        cloze += '~'.join(map(lambda s: s.to_cloze(), self.choices))

        return f"{{{cloze}}}"


@dataclass
class MultiTrueFalse(Question):
    """Question, which displays a column of statements, which can be answered in a binary fashion (usually True/False).

    :param scoring_method: Can either be `"subpoints"` for partial points or `"mtfonezero"` for all-or-nothing scoring.
    :type scoring_method: str

    :param shuffle: If `True`, statements will be displayed randomly.
    :type shuffle: bool

    :param options: A list answering options for each statement. Defaults to `["Wahr", "Falsch"]`.
    :type options: typing.List[str]
    """
    scoring_method: str = 'subpoints'
    shuffle: bool = True
    options: typing.List[str] = field(default_factory=lambda: ['Wahr', 'Falsch'])
    statements: typing.List[typing.Tuple[str, str, str]] = field(default_factory=list)

    def add_statement(self, text: str, value: str, feedback: str = ''):
        """Adds a True/False statement to the question.

        :param text: Text of the statement to be shown. Will be formatted in the format of the question itself.
        :type text: str

        :param value: The correct value out of the available choices in `self.options`.
        :type value: str

        :param feedback: Feedback to be shown after the answer was submitted. Will be rendered in the question format.
        :type feedback: str

        :raises ValueError: if `value` is not an answering option
        """
        if value not in self.options:
            raise ValueError(f'The supplied value "{value}" is not an available answering option!')

        self.statements.append((text, value, feedback))

    def to_xml(self) -> ET.Element:
        node = self.generate_common_xml('mtf')

        scoring = ET.SubElement(node, 'scoringmethod')
        ET.SubElement(scoring, 'text').text = self.scoring_method

        ET.SubElement(node, 'shuffleanswers').text = 'true' if self.shuffle else 'false'

        num_rows = len(self.statements)
        num_cols = len(self.options)

        ET.SubElement(node, 'numberofrows').text = str(num_rows)
        ET.SubElement(node, 'numberofcolumns').text = str(num_cols)

        for i, (text, _, feedback) in enumerate(self.statements, start=1):
            row = ET.SubElement(node, 'row')
            row.set('number', str(i))

            opt = ET.SubElement(row, 'optiontext')
            opt.set('format', self.format)
            ET.SubElement(opt, 'text').text = text

            fb = ET.SubElement(row, 'feedbacktext')
            fb.set('format', self.format)
            ET.SubElement(fb, 'text').text = feedback

        for i, opt in enumerate(self.options, start=1):
            col = ET.SubElement(node, 'column')
            col.set('number', str(i))

            resp = ET.SubElement(col, 'responsetext')
            resp.set('format', self.format)
            ET.SubElement(resp, 'text').text = opt

        for i, (_, correct, _) in enumerate(self.statements, start=1):
            for j, opt in enumerate(self.options, start=1):
                weight = ET.SubElement(node, 'weight')
                weight.set('rownumber', str(i))
                weight.set('columnnumber', str(j))

                ET.SubElement(weight, 'value').text = '1.000' if opt == correct else '0.000'

        return node


@dataclass
class Matching(Question):
    """Match drop-down menu items to the correct text.

    :param shuffle: If `True`, items will be shuffled.
    :type shuffle: bool
    """

    shuffle: bool = True
    pairs: typing.List[typing.Tuple[typing.Union[str, None], str]] = field(default_factory=list)

    def add_pair(self, text: typing.Union[str, None], answer: str):
        """Adds a matching pair to the question. If the `text` parameter is `None` a decoy option is added.

        :param text: The first element of the pair, displayed as normal text.
        :type text: str

        :param answer: The corresponding match to the `text` shown as drop-down menu item.
        :type text: str
        """
        self.pairs.append((text, answer))

    def to_xml(self) -> ET.Element:
        node = self.generate_common_xml('matching')

        for text, answer in self.pairs:
            subquestion = ET.SubElement(node, 'subquestion')
            subquestion.set('format', self.format)

            text = "" if text is None else text

            ET.SubElement(subquestion, 'text').text = text
            aw = ET.SubElement(subquestion, 'answer')
            ET.SubElement(aw, 'text').text = answer

        return node




#    Quizpy - Creating Moodle exams in Python
#    Copyright (C) 2021  Sebastian Bräuer
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

import xml.etree.ElementTree as ET
import typing

from dataclasses import dataclass, field
from .common import Question, InlineFile


# TODO Question Type ddwtos

@dataclass
class DragOption:
    """Represents an answering option box to be placed on a DropZone within a DDImage.

    :param text: The text on the box. This is treated as plain text without any markup.
    :type text: str

    :param group: Options can be grouped so that options with the same group number have the same shape that
                  is also represented in the DropZone, i.e. DragOptions can only be dragged on DropZones whos DragOption
                  is in the same group.
    :type group: str

    :param infinite: If `True` the option can be placed on an infinite number of DropZones.
    :type infinite: bool

    """
    text: str
    group: int = 1
    infinite: bool = False


@dataclass
class DropZone:
    """Represents a box in which a user can drop an answer on in a DDImage question.

    :param location: x,y coordinates of the drop zone relativ to the top left corner of the base image.
    :type location: typing.Tuple[int, int]

    :param choice: a reference to the `DragOption` that represents the correct answer for this field.

        .. attention:: choice must be a valid reference to a DragOption in DDImage.options

    :type choice: DragOption

    :param text:
        .. note:: TODO: Find out what this actually does. It is visible in the editor but seemingly not in the question?!
    """
    location: typing.Tuple[int, int]
    choice: DragOption
    text: str = ""


@dataclass
class DDImage(Question):
    """This question type allows a student to drop answers on custom-placed fields on a background image.

    :param base_image: an `InlineFile` instance that serves as the background to the question.

        .. attention:: Although this (for technical reasons) defaults to `None` you need to provide a background image
                       before you can call `to_xml()`!

    :type base_image: quizpy.InlineFile

    :param shuffle: If `True`, drag options will be sorted randomly each time the question is presented.
    """

    base_image: InlineFile = None  # Unfortunately,  this must be set with a default
    options: typing.List[DragOption] = field(default_factory=list)
    drops: typing.List[DropZone] = field(default_factory=list)
    shuffle: bool = False

    def add_option(self, text: str, group: int = 1, infinite: bool = False) -> DragOption:
        """Adds an DragOption to the DDImage question and returns the newly created instance.

        :param text: The text on the box. This is treated as plain text without any markup.
        :type text: str

        :param group: Options can be grouped so that options with the same group number have the same shape that
                      is also represented in the DropZone, i.e. DragOptions can only be dragged on DropZones whos DragOption
                      is in the same group.
        :type group: str

        :param infinite: If `True` the option can be placed on an infinite number of DropZones.
        :type infinite: bool
        """

        c = DragOption(text, group=group, infinite=infinite)
        self.options.append(c)
        return c

    def add_dropzone(self, location: typing.Tuple[int, int], correct_choice: DragOption, text: str = "") -> DropZone:
        """Adds an DropZone to the DDImage question and returns the newly created instance.

        :param location: x,y coordinates of the drop zone relativ to the top left corner of the base image.
        :type location: typing.Tuple[int, int]

        :param correct_choice: a reference to the `DragOption` that represents the correct answer for this field.

        .. attention:: choice must be a valid reference to a DragOption in DDImage.options

        :type choice: DragOption

        :param text:
             .. note:: TODO: Find out what this actually does. It is visible in the editor but seemingly not in the question?!

        """
        d = DropZone(location, correct_choice, text=text)
        self.drops.append(d)
        return d

    def to_xml(self) -> ET.Element:

        if self.base_image is None:
            raise ValueError("You need to set a base image!")

        node = self.generate_common_xml('ddimageortext')

        if self.shuffle:
            ET.SubElement(node, "shuffleanswers")

        node.append(self.base_image.to_xml())

        for i, opt in enumerate(self.options, start=1):
            drag = ET.SubElement(node, 'drag')
            ET.SubElement(drag, 'text').text = opt.text
            ET.SubElement(drag, 'draggroup').text = str(opt.group)
            ET.SubElement(drag, 'no').text = str(i)

            if opt.infinite:
                ET.SubElement(drag, 'infinite')

        for i, d in enumerate(self.drops, start=1):
            x, y = d.location
            drop = ET.SubElement(node, 'drop')
            ET.SubElement(drop, 'text')  # Empty sub-element
            ET.SubElement(drop, 'xleft').text = str(x)
            ET.SubElement(drop, 'ytop').text = str(y)
            ET.SubElement(drop, 'no').text = str(i)

            try:
                ET.SubElement(drop, 'choice').text = str(self.options.index(d.choice) + 1)
            except ValueError:
                raise ValueError("Choice for dropzone not found in drag options!")

        return node



#    Quizpy - Creating Moodle exams in Python
#    Copyright (C) 2021  Sebastian Bräuer
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

import typing
import base64
import abc
import xml.etree.ElementTree as ET
import os

from dataclasses import dataclass, field

@dataclass
class InlineFile:
    """Helper class to attach and reference files directly in the XML file.

    :param name: Name of the file including the file extension.
    :type name: str

    :param file: File-like object to read data from. Must be opened in binary mode.
    :type name: typing.BinaryIO
    """
    name: str
    file: typing.BinaryIO

    @classmethod
    def load(cls, path: str):
        """Open a file in binary-mode and use the file handle to instantiate an InlineFile."""
        f = open(path, 'rb')
        return cls(os.path.basename(path), f)

    def to_xml(self) -> ET.Element:
        """Reads the file and turns it into `file` node in XML. If the file is seekable it will be read from the start."""
        if self.file.seekable():
            self.file.seek(0)

        node = ET.Element('file')
        node.set('name', self.name)
        node.set('path', '/')
        node.set('encoding', 'base64')

        node.text = base64.standard_b64encode(self.file.read()).decode('utf-8')

        return node

    def link(self):
        """Returns a string that Moodle can use to reference the file. Useful for e.g. links to pictures.

        :rtype: str
        """
        return f'@@PLUGINFILE@@/{self.name}'

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.file.close()


@dataclass
class Question(abc.ABC):
    """Abstract base class for all question types.

    :param title: The question title
    :type title: str

    :param text: The full question text in its raw form.
    :type text: str

    :param default_points: The base number of points given for this question. This can be scaled in the actual quiz on moodle.
    :type default_points: float

    :param format: The format of the question text. Can either be `html` (default), `markdown`, `plain_text` or `moodle_auto_format`.
    :type format: str

    :param id: Apparently moodle supports some kind of questions identifiers. We have no idea, how.
    :type id: int

    :param tags: A set of tags to group questions by. Unfortunately, not yet fully supported in moodle.
    :type tags: typing.Iterable[str]

    :param penalty: A penalty factor for in case of retries. Only applies if retry is possible for the quiz.
    :type penalty: float

    :param hidden: A marker to prevent 'old' questions from being displayed in the category.
    :type hidden: bool

    :param general_feedback: Feedback that will be shown after the question was answered (but during the quiz!). Will be rendered in the same format as the question text.
    :type general_feedback: bool

    :param files: A list of `InlineFiles` that can be referenced in the question text (e.g. pictures).
    :type files: typing.List[InlineFile]
    """
    title: str
    text: str
    default_points: float

    format: str = 'html'
    id: int = None
    tags: typing.Iterable[str] = field(default_factory=list)
    penalty: float = 0.0
    hidden: bool = False
    general_feedback: str = ''

    files: typing.List[InlineFile] = field(default_factory=list)

    def generate_common_xml(self, qtype: str) -> ET.Element:
        """Generates XML node with attributes common to all question types.

        :param qtype: the XML-question type
        :type qtype: str
        """
        node = ET.Element('question')
        node.set('type', qtype)

        name = ET.SubElement(node, 'name')
        ET.SubElement(name, 'text').text = self.title

        text = ET.SubElement(node, 'questiontext')
        text.set('format', self.format)
        ET.SubElement(text, 'text').text = self.text

        for inline in self.files:
            text.append(inline.to_xml())

        feedback = ET.SubElement(node, 'generalfeedback')
        feedback.set('type', self.format)
        ET.SubElement(feedback, 'text').text = self.general_feedback

        ET.SubElement(node, 'defaultgrade').text = str(self.default_points)
        ET.SubElement(node, 'penalty').text = str(self.penalty)
        ET.SubElement(node, 'hidden').text = '1' if self.hidden else '0'
        ET.SubElement(node, 'idnumber').text = str(self.id) if self.id is not None else ''
        return node

    @abc.abstractmethod
    def to_xml(self) -> ET.Element:
        pass



@dataclass
class Category:
    """Represents a group of questions within one XML-file.

    :param name: The category name that will be displayed in Moodle.
    :type name: str

    :param description: A long formatted description of the content.
    :type description: str

    :param format: The format used to render the description.
        Can either be `html` (default), `markdown`, `plain_text` or `moodle_auto_format`.
    :type format: str
    """
    name: str
    description: str = ''
    questions: typing.List[Question] = field(default_factory=list)
    format: str = 'html'

    def to_xml(self) -> typing.List[ET.Element]:
        nodes = []

        category = ET.Element('question')  # Dummy question node
        category.set('type', 'category')

        name = ET.SubElement(category, 'category')
        ET.SubElement(name, 'text').text = self.name

        info = ET.SubElement(category, 'info')
        info.set('format', self.format)
        ET.SubElement(info, 'text').text = self.description

        nodes.append(category)

        for q in self.questions:
            nodes.append(q.to_xml())

        return nodes


@dataclass
class Quiz:
    categories: typing.List[Category] = field(default_factory=list)

    def to_xml(self) -> ET.Element:
        root = ET.Element('quiz')

        for c in self.categories:
            root.extend(c.to_xml())

        return root

    def add_category(self, name: str, description: str = '', format: str = 'html') -> Category:
        cat = Category(name, description=description, format=format)
        self.categories.append(cat)
        return cat

    def export(self, path):
        with open(path, 'wb') as f:
            tree = ET.ElementTree(self.to_xml())
            tree.write(f, encoding='UTF-8', xml_declaration=True)



#    Quizpy - Creating Moodle exams in Python
#    Copyright (C) 2021  Sebastian Bräuer
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

from .common import Quiz, Question, Category, InlineFile
from .multipe_choice import MultipleChoice, Matching, MultiTrueFalse, Choice
from .text import Essay, Cloze, Description, ShortAnswer
from .numerical import Numerical, NumericalAnswer, UnitBehaviour, UnitGradingType
from .drag import DragOption, DropZone, DDImage

__all__ = ['Quiz', 'Question', 'Category', 'InlineFile',
           'MultipleChoice', 'Matching', 'MultiTrueFalse', 'Choice',
           'Essay', 'Cloze', 'Description', 'ShortAnswer',
           'Numerical', 'NumericalAnswer', 'UnitBehaviour', 'UnitGradingType',
           'DragOption', 'DropZone', 'DDImage']



#    Quizpy - Creating Moodle exams in Python
#    Copyright (C) 2021  Sebastian Bräuer
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

import xml.etree.ElementTree as ET
import typing
import enum

from dataclasses import dataclass, field

from .common import Question
from .multipe_choice import Choice


class UnitGradingType(enum.Enum):
    NO_PENALTY = 0
    FRAC_RESPONSE_GRADE = 1
    FRAC_TOTAL_GRADE = 2


class UnitBehaviour(enum.Enum):
    Optional = 0
    Force = 1
    NoUnits = 3


@dataclass
class NumericalAnswer(Choice):
    """A Choice sub-class with added tolerance for Numerical questions.

    :param tolerance: The absolute tolerance that a students answer is allowed to deviate from the correct value.
        .. note:: In case you use units the tolerance is assumed to be in the base unit and is converted accordingly.
    """

    tolerance: float = 0.0

    def to_xml(self) -> ET.Element:
        answer = super().to_xml()
        ET.SubElement(answer, 'tolerance').text = str(self.tolerance)
        return answer

    def to_cloze(self) -> str:
        """Converts answer into cloze representation. Do not use directly. This used by ``Numerical.to_cloze()``."""
        if self.fraction == 100.0:
            s = f"={self.text}:{self.tolerance}#{self.feedback}"
        elif self.fraction == 0.0:
            s = f"{self.text}:{self.tolerance}#{self.feedback}"
        else:
            s = f"%{int(self.fraction)}%{self.text}:{self.tolerance}#{self.feedback}"
        return s


@dataclass
class Numerical(Question):
    """A question with a one-line answering option that should be a number with optional units.

    :param units_left: If ``True``, Moodle will try to parse the unit on the left side of the number e.g. $1.
    :type units_left: bool

    :param unit_grading_type: Determines the grading for using possibly the wrong unit:

        UnitGradingType.NO_PENALTY
            No penalty is applied for not using the base unit / no unit at all.

            .. caution:: This does not imply ``unit_behaviour = UnitBehaviour.Optional``.
                         If ``unit_behaviour = UnitBehaviour.Force`` Moodle just ignores
                         this setting on import and assumes a penalty.

        UnitGradingType.FRAC_RESPONSE_GRADE
            A fraction of the grade given for this particular answer is
            deducted from the total grade. Implies ``unit_behaviour = UnitBehaviour.Force``

        UnitGradingType.FRAC_TOTAL_GRADE
            A fraction of the total grade is deducted. Implies ``unit_behaviour = UnitBehaviour.Force``

        .. note:: The penalty is applied when

            1. Using a unit that has a multiplier other than 1.
            2. Using a unit that was not defined
            3. Not using a unit at all

    :type unit_grading_type: UnitGradingType

    :param unit_behaviour: Determines if and how units are used in the question:

        UnitBehaviour.Optional
            Units can be used but if no unit is provided the base unit is assumed

        UnitBehaviour.Force
            Units must be provided.

        UnitBehaviour.NoUnits
            No units are used in the question

    :type unit_grading_type: UnitBehaviour

    :param unit_penalty: The fractional penalty between 0 and 1.
        This is only used if ``unit_behaviour = UnitBehaviour.Force``

    :type unit_penalty: float
    """
    accepted_answers: typing.List[NumericalAnswer] = field(default_factory=list)
    units: typing.List[typing.Tuple[str, float]] = field(default_factory=list)
    units_left: bool = False
    unit_grading_type: UnitGradingType = UnitGradingType.NO_PENALTY
    unit_behaviour: UnitBehaviour = UnitBehaviour.NoUnits
    unit_penalty: float = 0.1

    # TODO: Eingabeformat wird im XML komplett vergessen

    def to_xml(self) -> ET.Element:
        node = self.generate_common_xml('numerical')

        for answer in self.accepted_answers:
            node.append(answer.to_xml())

        units = ET.SubElement(node, 'units')
        for unit, multiplier in self.units:
            u = ET.SubElement(units, 'unit')
            ET.SubElement(u, 'multiplier').text = str(multiplier)
            ET.SubElement(u, 'unit_name').text = unit

        ET.SubElement(node, 'unitgradingtype').text = str(self.unit_grading_type.value)
        ET.SubElement(node, 'unitsleft').text = '1' if self.units_left else '0'
        ET.SubElement(node, 'showunits').text = str(self.unit_behaviour.value)
        ET.SubElement(node, 'unitpenalty').text = str(self.unit_penalty)

        return node

    def to_cloze(self, weight: int):
        """Turns this question into its Cloze syntax representation to be used in a Cloze question text."""
        cloze = f'{weight}:NM:'
        cloze += '~'.join(map(lambda a: a.to_cloze(), self.accepted_answers))

        return f'{{{cloze}}}'

    def add_unit(self, unit: str, conversion_factor: float):
        """Convenience method to add a unit.

        :param unit: A string representing the unit.
            .. note:: Due to this bug <https://tracker.moodle.org/browse/MDL-69597> we strip the whitespaces before
            adding it to the list.

        :type unit: str

        :param conversion_factor: The factor that is applied before comparing it with the correct answer.

            .. note:: If you use units you should always add at least one option with ``conversion factor = 1.0`` so
                      that your unit will be accepted as the correct unit.
        """

        self.units.append((unit.strip(), conversion_factor))

    def add_answer(self, number: float, fraction: float, feedback: str = '', format: str = 'html', tolerance: float = 0.0) -> NumericalAnswer:
        """Convenience method to add a ``NumericalAnswer``
        :param number: The correct number (assume the base unit in case that units are used).
        :type number: float

        :param fraction: The fraction of the total points to be given for this answer as percentage between 0 and 100.
        :type fraction: float

        :param feedback: Specific feedback for choosing this answer option
        :type feedback: str

        :param format: format for rendering both answer text and feedback.
            Can either be `html` (default), `markdown`, `plain_text` or `moodle_auto_format`.
        """

        n = NumericalAnswer(f"{number}", fraction, feedback=feedback, format=format, tolerance=tolerance)
        self.accepted_answers.append(n)
        return n




#    Quizpy - Creating Moodle exams in Python
#    Copyright (C) 2021  Sebastian Bräuer
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

import xml.etree.ElementTree as ET
import typing

from dataclasses import dataclass, field

from .common import Question
from .multipe_choice import Choice


@dataclass
class ShortAnswer(Question):
    """Question type asking for a one-line answer.

    :param case_sensitive: If `True`, the given answer must match the solution exactly.
    :type case_sensitive: bool
    """
    case_sensitive: bool = False
    accepted_answers: typing.List[Choice] = field(default_factory=list)

    def add_answer(self, text: str, fraction: float, feedback: str = '', format: str = 'html') -> Choice:
        """Convenience method to add an answer to the question.

        The answer itself is the same `Choice` object that MultipleChoice questions use.

        :param text: The source text to be rendered for this answer option.
        :type text: str

        :param fraction: The fraction of the total points to be given for this answer as percentage between 0 and 100.
        :type fraction: float

        :param feedback: Specific feedback for choosing this answer option
        :type feedback: str

        :param format: format for rendering both answer text and feedback.
            Can either be `html` (default), `markdown`, `plain_text` or `moodle_auto_format`.

        :returns: The newly added Choice object.
        :rtype: quizpy.Choice
        """

        c = Choice(text, fraction, feedback, format)
        self.accepted_answers.append(c)
        return c

    def to_xml(self) -> ET.Element:
        node = self.generate_common_xml('shortanswer')
        ET.SubElement(node, 'usecase').text = '1' if self.case_sensitive else 0

        for answer in self.accepted_answers:
            node.append(answer.to_xml())

        return node

    def to_cloze(self, weight: int) -> str:
        """Renders a `ShortAnswer` question in Cloze format to be used as an inline question.

        :param weight: The weight of the question in the total sum of all inline questions.
        :type weight: int
        """
        cloze = f"{weight}:SA:" if not self.case_sensitive else f"{weight}:SAC:"

        cloze += '~'.join(map(lambda s: s.to_cloze(), self.accepted_answers))

        return f"{{{cloze}}}"


class Description(Question):
    """Dummy question type with no grading e.g. to establish some context for further questions."""

    def to_xml(self) -> ET.Element:
        node = self.generate_common_xml('description')
        return node


@dataclass
class Essay(Question):
    """Question with free-form multiline text answer that must be graded manually.

    .. note:: Moodle allows you to restrict file types of attachements. However, we do not fully understand how this
              works yet.

    :param grading_info: Text to be displayed for the grader (students will never see this)
        e.g. to establish rules for grading. Will be formatted in the question format.

    :type grading_info: str

    :param response_format: Defines how a student can provide an answer. Can be one of the following options:

        * "editor": Standard HTML editor (with possibility to add pictures)
        * "editorfilepicker": Same as editor but with file upload option
        * "plain": Unformatted plain text input.
        * "monospaced": Unformatted plain text input in monospace font.
        * "noinline": No text input field. This requires `attachements_required = True` and `max_attachments > 0`.

    :type response_format: str

    :param response_required: If `True`, a student must enter some text before submitting the answer.
    :type response_required: bool

    :param response_template: A formatted text that is automatically entered as a template for the student.
    :type response_template: str

    :param attachements_required: If `True`, a student must submit a file before submitting the answer.
    :type attachements_required: bool

    :param max_attachements: The maximum number of attachements a student can supply for an answer.
    :type max_attachements: int

    :param max_lines: The maximum number of lines of text an answer can have.
    :type max_lines: int
    """
    grading_info: str = ""
    response_format: str = "editor"
    response_required: bool = True
    response_template: str = ""
    attachements_required: bool = False
    max_attachements: int = 0
    max_lines: int = 15

    def to_xml(self) -> ET.Element:
        node = self.generate_common_xml('essay')

        ET.SubElement(node, 'responseformat').text = self.response_format
        ET.SubElement(node, 'responserequired').text = '1' if self.response_required else '0'
        ET.SubElement(node, 'responsefieldlines').text = str(self.max_lines)
        ET.SubElement(node, 'attachements').text = str(self.max_attachements)
        ET.SubElement(node, 'attachementsrequired').text = '1' if self.attachements_required else '0'

        grading_info = ET.SubElement(node, 'graderinfo')
        grading_info.set('format', self.format)
        ET.SubElement(grading_info, 'text').text = self.grading_info

        template = ET.SubElement(node, 'responsetemplate')
        template.set('format', self.format)
        ET.SubElement(template, 'text').text = self.response_template

        return node


class Cloze(Question):
    """The question text will be rendered using the custom Cloze syntax.
    See https://docs.moodle.org/310/en/Embedded_Answers_(Cloze)_question_type.

    You can also use the `to_cloze()` functions of the relevant question classes to generate the syntax for you.
    """

    def to_xml(self) -> ET.Element:
        node = self.generate_common_xml('cloze')
        return node
